#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <windows.h>

void (*cpuwrite)(uint16_t, uint8_t);
uint8_t (*cpuread)(uint16_t);

void InitMassStor( int slot );
void ShutdownMassStor( void );
unsigned char ReadMassStorIO( unsigned short Address );
void WriteMassStorIO( unsigned short Address, unsigned char Data );

off_t FileSize(int handle)
{
 off_t c, l;
 
 c=lseek(handle, 0, SEEK_CUR);
 l=lseek(handle, 0, SEEK_END);
   lseek(handle, c, SEEK_SET);
 
 return l;
}


/* NOTE: in theory ProDOS can support 4 volumes in slots 5 & 6. However,
   it never seems to perform STATUS or FORMAT calls for the 2 extra units.
   Hence, the mass-storage routines supply ProDOS with incorrect block
   numbers, and won't format correctly. So extra volumes are disabled for
   now. */

/* #define MS_DMA to create DMA (direct memory access) version
   #undef  MS_DMA to create slow I/O (but more standard) version
                  where the card does not access memory
   Make sure the correct version of MASSSTOR.ASM has been assembled. */
#define MS_DMA

/*
;Mass-storage device I/O
; $0 - Writing the ProDOS command here will perform that command.
;      It is important to setup the other locations first.
;      0=Status,1=Read,2=Write,3=Format
; $1 - Unit number $0-$3
; $2 - ProDOS buffer low byte
; $3 - ProDOS buffer high byte
; $4 - Block low byte
; $5 - Block high byte
; $6 - Data register
*/

#define PS_OK    0x00
#define PS_IOERR 0x27
#define PS_NODEV 0x28
#define PS_WRPRO 0x2b

int MassStorSlot;

unsigned short MaxBlock = 65535; //280;

char VolumeFN[ 4 ][ 80 ]; /* volume file names */
int VolumeFH[ 4 ];        /* volume file handles */
long VolumeSize[ 4 ];     /* size of volumes in bytes */
int WritePro[ 4 ];        /* if a volume is write-protected */
int ValidBlock[ 4 ];      /* if current block buffer contains valid data */
unsigned short int Block[ 4 ];          /* block # in buffer */
unsigned char Buffer[ 4 ][ 512 ];  /* 512 byte buffer for each unit */

char HDDROM[256]; /* no room in dapple.c, moved here -uso. */

//int hdd=8;

unsigned char Unit;            /* current unit */
unsigned short ProDOSBuf;       /* Apple memory location to ProDOS buffer */
unsigned char Byte[ 4 ];       /* current byte offset in buffer */
unsigned char DataReg;         /* current value of data register */
unsigned char Status;          /* current status of device */

#define STATUS_OFF 0
#define STATUS_RD  1
#define STATUS_WR  2

/* Updates light for current unit based on command */
/* cmd:0=nothing,1=reading,2=writing */
void UpdateMassStorStatus( int cmd )
{
 //hdd=(cmd?((cmd==1)?2:4):8);
}

void InitMassStor( int slot )
{
    int i;
    unsigned int attr;

    MassStorSlot = slot;

    for ( i = 0; i < 4; i++ )
    {
        Unit=i;
        UpdateMassStorStatus( STATUS_OFF );
        //attr = GetAttrib( VolumeFN[ i ] );
        //if ( attr == 0xffff )
        //{
            attr = 0;
        //}
        WritePro[ i ] = /*( attr & FA_RDONLY ) ? 1 :*/ 0;
        VolumeFH[ i ] = open( VolumeFN[ i ], /*WritePro[ i ] ? O_RDONLY :*/ O_RDWR|O_BINARY );
        if ( VolumeFH[ i ] > 0 )
        {
            VolumeSize[ i ] = FileSize( VolumeFH[ i ] );
        }
        else
        {
            VolumeSize[ i ] = 0L;
        }
        ValidBlock[ i ] = 0;
    }
    Status = 0;
    Unit = 0;
}

void ShutdownMassStor( void )
{
    int i;

    for ( i = 0; i < 4; i++ )
    {
        if ( VolumeFH[ i ] > 0 )
        {
            close( VolumeFH[ i ] );
            if ( VolumeSize[ i ] == 0L )
            {
                /* nothing in the file - might as well delete it */
                remove( VolumeFN[ i ] );
            }
        }
        VolumeFH[ i ] = 0;
    }
}


unsigned char ReadMassStorIO( unsigned short Address )
{
    unsigned char v;

    v = 0xff;
    switch ( Address & 0x0f )
    {
    case 0x01:
        v = Unit;
        break;
    case 0x02:
        #ifdef MS_DMA
        v = ProDOSBuf & 0xff;
        #else
        v = Byte[ Unit ] & 0xff;
        #endif
        break;
    case 0x03:
        #ifdef MS_DMA
        v = ProDOSBuf >> 8;
        #else
        v = Byte[ Unit ] >> 8;
        #endif
        break;
    case 0x04:
        v = Block[ Unit ] & 0xff;
        break;
    case 0x05:
        v = Block[ Unit ] >> 8;
        break;
    case 0x06:
        v = DataReg;
        break;
    case 0x07:
        v = Status;
        break;
    }
    return v;
}

/* this does not actually erase any data - it only attempts to create a
   volume of the size given in MaxBlock */
void FormatVolume( void )
{
    if ( WritePro[ Unit ] ) {
        Status = PS_WRPRO;
    }
    else {
        lseek( VolumeFH[ Unit ], 512L * (long)MaxBlock -1L, SEEK_SET );
        write( VolumeFH[ Unit ], Buffer[ Unit ], 1 );
        VolumeSize[ Unit ] = (long)MaxBlock * 512L;
    }
}

void ReadBlock( void )
{
    #ifdef MS_DMA
    int bufptr;
    #endif

    if ( !ValidBlock[ Unit ] )
    {
        /* get block from disk */
        if ( VolumeFH[ Unit ] < 1 )
        {
            Status = PS_NODEV;
        }
        else
        {
            lseek( VolumeFH[ Unit ], 512L * (long)Block[ Unit ], SEEK_SET );
            read( VolumeFH[ Unit ], Buffer[ Unit ], 512 );
            ValidBlock[ Unit ] = 1;
            Byte[ Unit ] = 0;
        }
    }
    if ( Status == PS_OK )
    {
        #ifdef MS_DMA
        for ( bufptr = 0; bufptr < 512; bufptr++ )
        {
            cpuwrite( ProDOSBuf + bufptr, Buffer[ Unit ][ bufptr ] );
        }
        #else
        DataReg = Buffer[ Unit ][ Byte[ Unit ] ];
        Byte[ Unit ]++;
        if ( Byte[ Unit ] == 0x200 )
        {
            Byte[ Unit ] = 0;
        }
        #endif
    }
}

void WriteBlock( void )
{
    #ifdef MS_DMA
    int bufptr;
    #endif

    #ifdef MS_DMA
    for ( bufptr = 0; bufptr < 512; bufptr++ )
    {
        Buffer[ Unit ][ bufptr ] = cpuread( ProDOSBuf + bufptr );
    }
    Byte[ Unit ] = 0;
    #else
    Buffer[ Unit ][ Byte[ Unit ] ] = DataReg;
    Byte[ Unit ]++;
    if ( Byte[ Unit ] == 0x200 )
    {
        Byte[ Unit ] = 0;
    }
    #endif
    /* If not a valid filehandle, then no device error */
    if ( VolumeFH[ Unit ] < 1 )
    {
        Status = PS_NODEV;
        return;
    }
    /* If write-protected, then write-protected error */
    if ( WritePro[ Unit ] )
    {
        Status = PS_WRPRO;
        return;
    }
    /* Only write when the buffer is full */
    if ( Byte[ Unit ] == 0 )
    {
        /* Try to seek to the blocks location in the file */
        if ( lseek( VolumeFH[ Unit ], 512L * (long)Block[ Unit ], SEEK_SET ) == -1L )
        {
            /* Error */
            Status = PS_IOERR;
            return;
        }
        /* If the file is newly created (zero-length) then it needs to be
           formatted first */
        if ( VolumeSize[ Unit ] == 0L )
        {
            /* Check if the volume should be formatted */
//          if ( ShouldFormat )
//          {
//              FormatVolume();
//          }
        }
        /* Everything appears to be OK up until now - write the buffer */
        if ( _write( VolumeFH[ Unit ], Buffer[ Unit ], 512 ) == 0 )
        {
            /* Error on write */
            Status = PS_IOERR;
        }
    }
}

void WriteMassStorIO( unsigned short Address, unsigned char Data )
{
    Status = PS_OK;
    switch ( Address & 0x0f )
    {
    case 0x00:
        switch ( Data )
        {
        case 0x00:
            /* Status */
            /* Set # blocks */
            Block[ Unit ] = ( VolumeSize[ Unit ] == 0L ) ?
                            MaxBlock : (unsigned short)( VolumeSize[ Unit ] / 512L );
            break;
        case 0x01:
            UpdateMassStorStatus( STATUS_RD );
            ReadBlock();
            break;
        case 0x02:
            UpdateMassStorStatus( STATUS_WR );
            WriteBlock();
            break;
        case 0x03:
            UpdateMassStorStatus( STATUS_WR );
            FormatVolume();
            break;
        }
        UpdateMassStorStatus( STATUS_OFF );
        break;
    case 0x01:
        Unit = Data;
        break;
    case 0x02:
        #ifdef MS_DMA
        ProDOSBuf = ( ProDOSBuf & 0xff00 ) | Data;
        #else
        Byte[ Unit ] = ( Byte[ Unit ] & 0xff00 ) | Data;
        #endif
        break;
    case 0x03:
        #ifdef MS_DMA
        ProDOSBuf = ( ProDOSBuf & 0x00ff ) | ( Data << 8 );
        #else
        Byte[ Unit ] = ( Byte[ Unit] & 0x00ff ) | ( Data << 8 );
        #endif
        break;
    case 0x04:
        Block[ Unit ] = ( Block[ Unit ] & 0xff00 ) | Data;
        ValidBlock[ Unit ] = 0;
        break;
    case 0x05:
        Block[ Unit ] = ( Block[ Unit ] & 0x00ff ) | ( Data << 8 );
        ValidBlock[ Unit ] = 0;
        break;
    case 0x06:
        DataReg = Data;
        break;
  }
}

/*
 * The MoDapple interface
 */

static char massstor_rom[256]={
 0xA2, 0x20, 0xA0, 0x00, 0xA2, 0x03, 0xA2, 0x3C,
 0x08, 0x78, 0x20, 0x58, 0xFF, 0xBA, 0xBD, 0x00,
 0x01, 0x28, 0xAA, 0x0A, 0x0A, 0x0A, 0x0A, 0x85,
 0x2B, 0x85, 0x43, 0xA9, 0x01, 0x85, 0x42, 0xA9,
 0x00, 0x85, 0x44, 0xA9, 0x08, 0x85, 0x45, 0xA9,
 0x00, 0x85, 0x46, 0x85, 0x47, 0x8A, 0x48, 0xA9,
 0x37, 0x48, 0x8A, 0x48, 0xA9, 0x3C, 0x48, 0x60,
 0xA6, 0x2B, 0x4C, 0x01, 0x08, 0xA5, 0xFF, 0x48,
 0xA9, 0x60, 0x85, 0xFF, 0x08, 0x78, 0x20, 0xFF,
 0x00, 0xBA, 0xBD, 0x00, 0x01, 0x28, 0x0A, 0x0A,
 0x0A, 0x0A, 0xAA, 0xA9, 0x00, 0x9D, 0x81, 0xC0,
 0xA5, 0x43, 0x29, 0x70, 0x85, 0xFF, 0xE4, 0xFF,
 0xF0, 0x21, 0xE0, 0x50, 0xF0, 0x0B, 0xE0, 0x60,
 0xF0, 0x0D, 0x68, 0x85, 0xFF, 0x38, 0xA9, 0x28,
 0x60, 0xC9, 0x10, 0xD0, 0xF5, 0xF0, 0x04, 0xC9,
 0x20, 0xD0, 0xEF, 0xBD, 0x81, 0xC0, 0x09, 0x02,
 0x9D, 0x81, 0xC0, 0xA5, 0x43, 0x0A, 0x90, 0x08,
 0xA9, 0x01, 0x1D, 0x81, 0xC0, 0x9D, 0x81, 0xC0,
 0xA5, 0x44, 0x9D, 0x82, 0xC0, 0xA5, 0x45, 0x9D,
 0x83, 0xC0, 0xA5, 0x46, 0x9D, 0x84, 0xC0, 0xA5,
 0x47, 0x9D, 0x85, 0xC0, 0xA5, 0x42, 0x9D, 0x80,
 0xC0, 0x68, 0x85, 0xFF, 0x18, 0xBD, 0x87, 0xC0,
 0xD0, 0x0F, 0xA5, 0x42, 0xD0, 0x08, 0xBD, 0x85,
 0xC0, 0xA8, 0xBD, 0x84, 0xC0, 0xAA, 0xA9, 0x00,
 0x60, 0x38, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9F, 0x3D
};

__declspec(dllexport) uint32_t mod_type=0x544F4C53;
__declspec(dllexport) uint8_t *mod_name="Mass Storage Interface";
__declspec(dllexport) uint8_t *mod_pref="massstor";
__declspec(dllexport) uint8_t *mod_vers="0.03";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas\nOriginal code by Andrew Gregory";

__declspec(dllexport) uint8_t  *massstor_preload_name="";
__declspec(dllexport) uint16_t massstor_preload_size=0;
__declspec(dllexport) uint8_t  *massstor_preload_ptr=0;

__declspec(dllexport) void massstor_write16 (uint16_t a, uint8_t v)
{
 return WriteMassStorIO(a&0x0F,v);
}
__declspec(dllexport) void massstor_write256 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) void massstor_write2048 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) uint8_t massstor_read16 (uint16_t a)
{
 return ReadMassStorIO(a&0x0F);
}

__declspec(dllexport) uint8_t massstor_read256 (uint16_t a)
{
 return massstor_rom[a&0x00FF];
}

__declspec(dllexport) uint8_t massstor_read2048 (uint16_t a)
{
 return 0;
}

__declspec(dllexport) int16_t massstor_override_wr (uint16_t a, uint8_t v)
{
 return -1; /* we don't do this. */
}

__declspec(dllexport) int16_t massstor_override_rd (uint16_t a)
{
 return -1; /* we don't do this. */
}

__declspec(dllexport) int massstor_reset (void)
{
 return 0;
}

__declspec(dllexport) void massstor_dmainit(void (*v_cpuwrite)(uint16_t, uint8_t),
                                            uint8_t (*v_cpuread)(uint16_t))
{
 cpuwrite=v_cpuwrite;
 cpuread=v_cpuread;
}

static char inifile[1024];

__declspec(dllexport) int massstor_init (void)
{
 int l;
 char modbuf[256];
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 l=GetPrivateProfileString("mod_massstor","Drive1","disk7a.hdv",modbuf,256,inifile);
 if (l) strcpy(VolumeFN[0],modbuf);
 l=GetPrivateProfileString("mod_massstor","Drive2","disk7b.hdv",modbuf,256,inifile);
 if (l) strcpy(VolumeFN[1],modbuf);

 InitMassStor(7);
 return 6502;
}

__declspec(dllexport) int massstor_deinit (void)
{
 int l;
 ShutdownMassStor();
 l=WritePrivateProfileString("mod_massstor","Drive1",VolumeFN[0],inifile);
 l=WritePrivateProfileString("mod_massstor","Drive2",VolumeFN[1],inifile);
 
 return 0;
}

__declspec(dllexport) void massstor_config (void)
{
}
