/*
 * A basic video and keyboard driver for MoDapple.
 * 
 * * mod_type should be 0x54545950 (TTYP)
 * * mod_name should be the display name of the module
 * * mod_pref should be the name prefixed to module functions
 */

#include <curses.h>
#include <stdint.h>
#include <windows.h>

__declspec(dllexport) uint32_t mod_type=0x50595454;
__declspec(dllexport) uint8_t *mod_name="Curses";
__declspec(dllexport) uint8_t *mod_pref="ttycurses";
__declspec(dllexport) uint8_t *mod_vers="0.02";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas";

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
} IIE;

static IIE *iie;

static uint8_t vramt[2][1024];
static uint8_t vramg[2][8192];

static int nogx, split, page2, hires, ttyok;
__declspec(dllexport) uint8_t ttycurses_ttykey;

static int grcolors[16]={
 COLOR_BLACK, COLOR_RED, COLOR_BLUE, COLOR_MAGENTA, 
 COLOR_GREEN, COLOR_WHITE, COLOR_BLUE, COLOR_BLUE,
 COLOR_RED, COLOR_YELLOW, COLOR_WHITE, COLOR_MAGENTA,
 COLOR_GREEN, COLOR_YELLOW, COLOR_GREEN, COLOR_WHITE
};

#define A_AII_FLASH   A_REVERSE
#define A_AII_INVERSE A_REVERSE
#define A_AII_BAR     "\xDF"

static void mark (uint16_t addr)
{
  uint8_t eights, ones, x, y, value;
  
  if (addr>0x03FF) return;
  
  value=vramt[page2][addr];
  
  eights=(addr&0x7F)/40;
  ones=(addr&0x03FF)>>7;
  x=(addr&0x7F)%40;
  y=(eights<<3)|ones;
  
  if ((x<40)&&(y<24)) 
  {
   if ((nogx)||(split&&(y>19)))
   {
    if (value<0x20) mvaddch(y,x,value|0x40|A_AII_INVERSE);
     else if (value<0x40) mvaddch(y,x,value|A_AII_INVERSE);
     else if (value<0x60) mvaddch(y,x,value|A_AII_FLASH);
     else if (value<0x80) mvaddch(y,x,(value&0x3F)|A_AII_FLASH);
     else if (value<0xA0) mvaddch(y,x,(value&0x1F)|0x40);
     else if (value==0xFF) mvaddch(y,x,ACS_CKBOARD);
     else mvaddch(y,x,value&0x7F);
   }
   else
   {
    if (!value)
    {
     attron(COLOR_BLACK);
     mvprintw(y,x," ");
     attroff(COLOR_BLACK);
    }
    else
    {
     attron(COLOR_PAIR(value));
     mvprintw(y,x,A_AII_BAR);
     attroff(COLOR_PAIR(value));
    }
   }
  }
}

__declspec(dllexport) void ttycurses_ttywrite (uint16_t a, uint8_t v)
{
 if (a<1024)             return;
 if ((a>3071)&&(a<8192)) return;
 if (a>24575)            return;
 if (a<2048)             {vramt[0][a-1024]=v;  if(!page2) mark(a-1024); return;}
 if (a<3072)             {vramt[1][a-2048]=v;  if(page2)  mark(a-2048); return;}
 if (a<16384)            {vramg[0][a-8192]=v;                           return;}
 if (a<24576)            {vramg[1][a-16384]=v;                          return;}
}

__declspec(dllexport) uint8_t ttycurses_ttyread (uint16_t a)
{
 if (a<1024)             return 0;
 if ((a>3071)&&(a<8192)) return 0;
 if (a>24575)            return 0;
 if (a<2048)             return vramt[0][a-1024];
 if (a<3072)             return vramt[1][a-2048];
 if (a<16384)            return vramg[0][a-8192];
 if (a<24576)            return vramg[1][a-16384];
}

static void ttycurses_do (uint8_t a)
{
 uint16_t t;
 
 switch (a&0x000F)
 {
  case 0:
   nogx=0;
   break;
  case 1:
   nogx=1;
   break;
  case 2:
   split=0;
   break;
  case 3:
   split=1;
   break;
  case 4:
   page2=0;
   break;
  case 5:
   page2=1;
   break;
  case 6:
   hires=0;
   break;
  case 7:
   hires=1;
   break;
 }
 
 for (t=0; t<1024; t++)
 {
  mark(t);
 }
}

__declspec(dllexport) void ttycurses_ttyout (uint8_t b, uint8_t v)
{
 uint16_t a;
 
 a=0xC000|b;
 
 if ((a&0xFFF0)==0xC010)
 {
  ttycurses_ttykey &= 0x7F;
  return;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttycurses_do(a&0x000F);
  return;
 }
}

__declspec(dllexport) uint8_t ttycurses_ttyin (uint8_t b)
{
 uint16_t a;
 
 a=0xC000|b;
 if ((a&0xFFF0)==0xC000)
  return ttycurses_ttykey;
 if ((a&0xFFF0)==0xC010)
 {
  ttycurses_ttykey &= 0x7F;
  return ttycurses_ttykey;
 }
 if ((a&0xFFF0)==0xC030)
 {
  
 }
 if ((a&0xFFF0)==0xC050)
 {
  ttycurses_do(a&0x000F);
  return 0;
 }
 return 0;
}

__declspec(dllexport) void ttycurses_ttydeinit(void)
{
 if (!ttyok) return;
 clear();
 refresh();
 curs_set(1);
 endwin();
}


__declspec(dllexport) void ttycurses_ttyinit(IIE *main_iie)
{
 int t;

 iie=main_iie;
 if (iie->iiemode)
 {
  iie->alt[0]=iie->alt[1]=0;
 }
 
 initscr();
 raw();
 keypad(stdscr, TRUE);
 noecho();
 curs_set(0);
 nodelay(stdscr, TRUE);
 start_color();
 use_default_colors();
 for (t=0; t<256; t++)
 {
  init_pair(t, grcolors[t&0x0F], grcolors[(t&0xF0)>>4]);
 }
 ttyok=1;
}

__declspec(dllexport) int ttycurses_ttyloop(void)
{
 int k;

 refresh();
 
 k=getch();
 if (iie->iiemode)
 {
  if (k==KEY_PPAGE) iie->alt[0]!=iie->alt[0];
  if (k==KEY_NPAGE) iie->alt[1]!=iie->alt[1];
 }
 if (k!=ERR)
 {
  switch (k)
  {
   case KEY_BACKSPACE:
    ttycurses_ttykey=0xFF;
    return 0;
   case KEY_UP:
    ttycurses_ttykey=0x8B;
    return 0;
   case KEY_DOWN:
    ttycurses_ttykey=0x8A;
    return 0;
   case KEY_LEFT:
    ttycurses_ttykey=0x88;
    return 0;
   case KEY_RIGHT:
    ttycurses_ttykey=0x95;
    return 0;
   case '\n':
    ttycurses_ttykey=0x8D;
    return 0;

    /* Slot config */
   case KEY_F(1): return 11;
   case KEY_F(2): return 12;
   case KEY_F(3): return 13;
   case KEY_F(4): return 14;
   case KEY_F(5): return 15;
   case KEY_F(6): return 16;
   case KEY_F(7): return 17;
   case KEY_F(8): return 10;

   case KEY_F(10):
   case KEY_F(40):
    return -1;
   case KEY_F(36):
   case KEY_BREAK:
    return 1;
   default:
    if (k<0x80)
    {
     /* flip case */
     if ((k>='A')&&(k<='Z')) k|=32;
      else if ((k>='a')&&(k<='z')) k&=223;
     ttycurses_ttykey=k|0x80;
    }
  }
 }
 return 0;
}
