#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

static char *verstr=
  "MoDapple  Version 0.15\n"
  "Copyright 2016 Steve Nickolas\n"
  "Other copyrights apply to plugin code\n";
static char *ESTR_NOTMOD ="%s is not a module\n",
            *ESTR_NOFUNC ="Could not find a '%s' function in %s\n",
            *ESTR_FSIZE  ="%s: Wrong file size (%lu expected, %lu found)\n",
            *ESTR_SHORT  ="%s: Short read (%lu expected, %lu found\n",
            *ESTR_INITMOD="Could not initialize module %s\n",
            *ESTR_SLOT   ="Slot %d",
            *ESTR_IS     =" is %s Version %s\n%s\n",
            *ESTR_LINK   ="Could not link %s to slot %d\n",
            *ESTR_NOTTY  ="No console loaded\n",
            *ESTR_NOCPU  ="No CPU loaded\n",
            *ESTR_NOBIOS ="No BIOS loaded\n",
            *ESTR_BADSLOT="Invalid slot %d\n",
            *ESTR_OPEN   ="Could not open module %s\n",
            *ESTR_ISNOT  ="%s is not a ",
            *ESTR_NOTSLOT="slot module\n",
            *ESTR_NOTTTY ="console module\n",
            *ESTR_NOTCPU ="CPU module\n",
            *ESTR_CPU    ="CPU",
            *ESTR_TTY    ="Console";

static char *STRING_UNDERSCORE_STRING="%s_%s";
static char *MOD_TYPE="mod_type",
            *MOD_NAME="mod_name",
            *MOD_PREF="mod_pref",
            *MOD_VERS="mod_vers",
            *MOD_COPR="mod_copr",
            *INI_SYSTEM   ="System",
            *INI_SLOTCARDS="SlotCards",
            *INI_BIOS     ="BIOS",
            *INI_CONSOLE  ="Console",
            *INI_CPU      ="CPU",
            *INI_DEF_ROM  ="apple2p.rom";

static char *ARBY="rb"; /* for reading files */

typedef struct
{
 HMODULE  dll;
 uint32_t *mod_type;
 uint8_t  **mod_name;
 uint8_t  **mod_pref;
 uint8_t  **mod_vers;
 uint8_t  **mod_copr;
 uint8_t  **preload_name;
 uint16_t *preload_size;
 uint8_t  **preload_ptr;
 void     (*write16)     (uint16_t a, uint8_t v);
 void     (*write256)    (uint16_t a, uint8_t v);
 void     (*write2048)   (uint16_t a, uint8_t v);
 uint8_t  (*read16)      (uint16_t a);
 uint8_t  (*read256)     (uint16_t a);
 uint8_t  (*read2048)    (uint16_t a);
 int16_t  (*override_rd) (uint16_t a);
 int16_t  (*override_wr) (uint16_t a, uint8_t v);
 int      (*reset)       (void);
 
 /*
  * Formerly, init() was a boolean.
  * For plugins intended for MoDapple 0.09 and later, if init() returns "6502"
  * then it needs pointers to cpuread() and cpuwrite().  The function to take
  * these pointers is only linked if it is required.
  */
 int      (*init)        (void);
 int      (*deinit)      (void);
 
 void     (*dmainit)     (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t));

 void     (*config)      (void);
} SLOT;

SLOT slot[8];

HINSTANCE hinstance;

typedef struct {
 uint8_t iiemode; /* emulating a //e */
 uint8_t store80,
         rdaux,
         wraux,
         lcaux,
         cxoverride,
         c3override,
         txt80,
         dblgrx,
         mostxt,
         iou,
         alt[2];
 uint8_t lcread,
         lcwrite,
         lcalt;
} IIE;

IIE iie;



typedef struct {
 uint8_t nogx,
         split,
         page2,
         hires;
} VID;

VID vid;

uint8_t  ram[131072], rom[16384];

uint8_t  fnstr[256];

/* Data pointers and functinos for TTY driver */
uint32_t *tty_mod_type;
uint8_t  **tty_mod_name;
uint8_t  **tty_mod_pref;
uint8_t  **tty_mod_vers;
uint8_t  **tty_mod_copr;
uint8_t  *ttykey;
void     (*ttywrite)   (uint16_t a, uint8_t v);
uint8_t  (*ttyread)    (uint16_t a);
void     (*ttyout)     (uint8_t a, uint8_t v);
uint8_t  (*ttyin)      (uint8_t a);
void     (*ttydeinit)  (void);
void     (*ttyinit)    (IIE*);
int      (*ttyloop)    (void);

/* Data pointers and functinos for CPU driver */
uint32_t *cpu_mod_type;
uint8_t  **cpu_mod_name;
uint8_t  **cpu_mod_pref;
uint8_t  **cpu_mod_vers;
uint8_t  **cpu_mod_copr;
void     (*cpureset)   (void);
void     (*cpurun)     (void);
void     (*cpuinit)    (void (*v_cpuwrite)(uint16_t, uint8_t), uint8_t (*v_cpuread)(uint16_t), int8_t (*v_cpuloop)(void));
void     (*cpuseta)    (uint8_t val);
void     (*cpusetx)    (uint8_t val);
void     (*cpusety)    (uint8_t val);
void     (*cpusets)    (uint8_t val);
void     (*cpusetp)    (uint8_t val);
void     (*cpusetpc)   (uint16_t val);
void     (*cpusetrate) (int val);
uint8_t  (*cpugeta)    (void);
uint8_t  (*cpugetx)    (void);
uint8_t  (*cpugety)    (void);
uint8_t  (*cpugets)    (void);
uint8_t  (*cpugetp)    (void);
uint16_t (*cpugetpc)   (void);
int      (*cpugetrate) (void);

HMODULE mod_tty, mod_cpu;

void canlibs (void)
{
 if (mod_tty) FreeLibrary(mod_tty);
 if (mod_cpu) FreeLibrary(mod_cpu);
}

void notamod (char *filename)
{
 canlibs();
 fprintf (stderr, ESTR_ISNOT, filename);
 fprintf (stderr, ESTR_NOTMOD);
 exit(1);
}

void nofunc (char *filename, char *ident)
{
 canlibs();
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 exit(1);
}

int nofunc2 (int slotno, char *filename, char *ident)
{
 FreeLibrary(slot[slotno].dll);
 slot[slotno].dll=0;
 fprintf (stderr, ESTR_NOFUNC, ident, filename);
 return -1;
}

#define TFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*tty_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_tty,fnstr))) nofunc(ttymod, b)
#define CFUNCTINO(a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*cpu_mod_pref,b); if (!(a=(void*)GetProcAddress(mod_cpu,fnstr))) nofunc(cpumod, b)
#define SFUNCTINO(s,a,b) sprintf (fnstr,STRING_UNDERSCORE_STRING,*(slot[s].mod_pref),b); if (!(slot[s].a=(void*)GetProcAddress(slot[s].dll,fnstr))) return nofunc2(s, module, b)

int bload (char *filename, int expect, uint8_t *pointer)
{
 FILE *file;
 long l;
 
 file=fopen(filename,ARBY);
 if (!file)
 {
  perror(filename);
  return -1;
 }
 
 fseek(file,0,SEEK_END);
 l=ftell(file);
 fseek(file,0,SEEK_SET);
 if (l!=expect)
 {
  fclose(file);
  fprintf (stderr, ESTR_FSIZE, filename, expect, l);
  return -1;
 }
 l=fread(pointer,1,expect,file);
 fclose(file);
 if (l<expect)
 {
  fprintf(stderr, ESTR_SHORT, filename, expect, l);
  return -1;
 }
 
 return 0;
}

/*****************************************************************************/
#if 0
/*
 * Start of a joystick driver.  XXX: This code should be in a DLL.
 * XXX: How is the xy info sent to the ][?
 */

static JOYINFO joyinfo;
static unsigned int nojoys, whichjoy;
static BOOL idjoy1, idjoy2;
static int gotstick;

typedef struct
{
 int8_t x;
 int8_t y;
 uint8_t button[2];
} AIIJOY;

static AIIJOY aiijoy;

void joyread (void)
{
 joyGetPos(whichjoy,&joyinfo);
 
 aiijoy.x=(joyinfo.wXpos>>8)-127;
 aiijoy.y=(joyinfo.wYpos>>8)-127;
 aiijoy.button[0]=(joyinfo.wButtons&JOY_BUTTON1)?0x00:0x80;
 aiijoy.button[1]=(joyinfo.wButtons&JOY_BUTTON2)?0x00:0x80;
}

void joyinit (void)
{
 gotstick=0;
 
 if (!(nojoys=joyGetNumDevs())) return;
 
 idjoy1=(joyGetPos(JOYSTICKID1,&joyinfo)!=JOYERR_UNPLUGGED);
 idjoy2=((nojoys==2)&&(joyGetPos(JOYSTICKID2,&joyinfo)!=JOYERR_UNPLUGGED));
 if (idjoy1||idjoy2)
 {
  gotstick=1;
  whichjoy=idjoy1?JOYSTICKID1:JOYSTICKID2;
 }
}
#endif
/*****************************************************************************/

static int fast6, fast6on;

static int holdc8;

/*
 * The ][+'s language card is a plugin, so that it can be added and removed at
 * will.  The //e and later MUST have this feature, and so a second language
 * card is hardwired into the shell for this support.
 */
static void iielc (uint16_t a)
{
 switch (a&0x0F)
 {
  case 0x0:
  case 0x4:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x1:
  case 0x5:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0x2:
  case 0x6:
   iie.lcalt=0x80;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0x3:
  case 0x7:
   iie.lcalt=0x80;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
   
  case 0x8:
  case 0xC:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x00;
   break;
  case 0x9:
  case 0xD:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x80;
   break;
  case 0xA:
  case 0xE:
   iie.lcalt=0x00;
   iie.lcread=0x00;
   iie.lcwrite=0x00;
   break;
  case 0xB:
  case 0xF:
   iie.lcalt=0x00;
   iie.lcread=0x80;
   iie.lcwrite=0x80;
   break;
 }
}

void cpuwrite(uint16_t a, uint8_t v)
{
 int t, o;
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC000))
 {
  switch (a&0x000F)
  {
   case 0x0:
    iie.store80=0x00;
    break;
   case 0x1:
    iie.store80=0x80;
    break;
   case 0x2:
    iie.rdaux=0x00;
    break;
   case 0x3:
    iie.rdaux=0x80;
    break;
   case 0x4:
    iie.wraux=0x00;
    break;
   case 0x5:
    iie.wraux=0x80;
    break;
   case 0x6:
    iie.cxoverride=0x00;
    break;
   case 0x7:
    iie.cxoverride=0x80;
    break;
   case 0x8:
    iie.lcaux=0x00;
    break;
   case 0x9:
    iie.lcaux=0x80;
    break;
   case 0xA:
    iie.c3override=0x00;
    break;
   case 0xB:
    iie.c3override=0x80;
    break;
   case 0xC:
    iie.txt80=0x00;
    break;
   case 0xD:
    iie.txt80=0x80;
    break;
   case 0xE:
    iie.mostxt=0x00;
    break;
   case 0xF:
    iie.mostxt=0x80;
    break;
  }
 }

 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_wr(a, v);
   if (!o) return;
  }
 }
 
 if ((iie.iiemode)&&((a&0xFFF0)==0xC080))
 {
  iielc(a&0xF); 
  return;  
 }
 
 if ((a<0x200)&&(iie.iiemode)&&(iie.lcaux))
 {
  ram[65536+a]=v;
  return;
 }
 
 if ((a>1023)&&(a<3072))  {ttywrite(a,v);}
 if ((a>8191)&&(a<24576)) {ttywrite(a,v);}
 if (a<49152)             
 {
  ram[a]=v; 
  return;
 }
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   slot[s].write16(a,v);
   return;
  }
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  s=a&0x0700;
  s>>=8;
  holdc8=s;
  if (slot[s].dll)
  {
   slot[s].write256(a,v);
   return;
  }
 }

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  if ((iie.iiemode)&&(iie.cxoverride)) return;
  if (!slot[holdc8].dll) return;
  slot[holdc8].write2048(a-0xC800,v);
  if (a==0xCFFF) holdc8=0;
  return;
 }

 if (a>=0xD000)
 {
  if ((iie.iiemode)&&(iie.lcwrite))
  {
   if (a<0xE000)
   {
    if (iie.lcalt)
    {
     if (iie.lcaux) {ram[65536+(a-0x1000)]=v; return;}
     
     ram[a-0x1000]=v;
    }
   }
   if (iie.lcaux) {ram[65536+a]=v; return;}
   ram[a]=v;
  }
  return;
 }
 
 if ((a&0xFFF0)==0xC000)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC010)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC030)  {ttyout(a,v); return;}
 if ((a&0xFFF0)==0xC050)  {ttyout(a,v); return;}
 
 if (iie.iiemode)
 {
  if (a==0xC07E) {iie.iou=0x00; return;}
  if (a==0xC07F) {iie.iou=0x80; return;}
 }
}

/* User level scanline; XXX might not be in synch with the video plugin */
uint16_t ulscanline;

uint8_t cpuread(uint16_t a)
{
 int t, o;
 
 if ((a==0xC0E8)&&(fast6)) fast6on=0;
 if ((a==0xC0E9)&&(fast6)) fast6on=1;

 for (t=7; t>=(iie.iiemode?1:0); t--)
 {
  if (slot[t].dll)
  {
   o=slot[t].override_rd(a);
   if (o>=0) 
   {
    return o;
   }
  }
 }

 if ((iie.iiemode)&&((a&0xFFF0)==0xC080)) 
 {
  iielc(a&0xF); 
  return 0;
 }

 if ((a<0x0200)&&(iie.iiemode)&&(iie.lcaux)) return ram[65536+a];
 
 if ((a>1023)&&(a<3072))  return ttyread(a);
 if ((a>8191)&&(a<24576)) return ttyread(a);
 if (a<49152)             return ram[a];
 
 if ((a&0xFF80)==0xC080)
 {
  int s;
  s=a&0x0070;
  s>>=4;
  if (slot[s].dll)
  {
   return slot[s].read16(a);
  }
  else return 0;
 }
 if (((a&0xF800)==0xC000)&&(a>=0xC100))
 {
  int s;
  
  if ((iie.iiemode)&&(iie.cxoverride)) return rom[a-0xC000];
  
  s=a&0x0700;
  s>>=8;
  if ((s==3)&&(iie.iiemode)&&(iie.c3override))
  {
   return rom[a-0xC000];
  }
  holdc8=s;
  if (slot[s].dll)
  {
   return slot[s].read256(a);
  }
  return rand()%256; /* well, it does the job */
 }

 if (((a&0xF800)==0xC800)&&(iie.iiemode)&&(iie.cxoverride)) return rom[a-0xC000];

 if (((a&0xF800)==0xC800)&&holdc8)
 {
  uint8_t c;

  if (!slot[holdc8].dll) return 255;
  c=slot[holdc8].read2048(a-0xC800);
  if (a==0xCFFF) holdc8=0;
  return c;
 }

 if (a>=0xD000)
 {
  if (iie.iiemode)
  {
   if (iie.lcread)
   {
    if (a<0xE000)
    {
     if (iie.lcalt) return ram[a-0x1000];
    }
    return ram[a];
   }
   return rom[a-0xC000];
  }
  return rom[a-0xD000];
 }
 
 if ((a&0xFFF0)==0xC000)  return ttyin(a);
 if ((a&0xFFF0)==0xC010)
 {
  if (iie.iiemode)
  {
   uint8_t k;
   
   if (!(a&0x0F)) k=ttyin(a)&0x7F; else k=ttyin(0)&0x7F;
   switch (a&0x0F)
   {
    case 0x0:
     break;
    case 0x1:
     k|=iie.lcalt;
     break;
    case 0x2:
     k|=iie.lcread;
     break;
    case 0x3:
     k|=iie.rdaux;
     break;
    case 0x4:
     k|=iie.wraux;
     break;
    case 0x5:
     k|=iie.cxoverride;
     break;
    case 0x6:
     k|=iie.lcaux;
     break;
    case 0x7:
     k|=iie.c3override;
     break;
    case 0x8:
     k|=iie.store80;
     break;
    case 0x9:
     k|=(ulscanline<192)?0x80:0x00;
     break;
    case 0xA:
     /* XXX: TTY flags need to be carried over from the video plugin. This is nogx */
     break;
    case 0xB:
     /* XXX: TTY flags need to be carried over from the video plugin. This is split */
     break;
    case 0xC:
     /* XXX: TTY flags need to be carried over from the video plugin. This is page2 */
     break;
    case 0xD:
     /* XXX: TTY flags need to be carried over from the video plugin. This is hires */
     break;
    case 0xE:
     k|=iie.mostxt;
     break;
    case 0xF:
     k|=iie.txt80;
     break;
   }
   return k;
  }
  return ttyin(a);
 }
 if ((a&0xFFF0)==0xC030)  return ttyin(a);
 if ((a&0xFFF0)==0xC050)  return ttyin(a);

 if (iie.iiemode)
 {
  if ((a==0xC061)||(a==0xC069)) return iie.alt[0];
  if ((a==0xC062)||(a==0xC06A)) return iie.alt[1];
  
  if (a==0xC07E) {return iie.iou;}
 }
 
 return 0;
}

static int die;

void reset_all(void)
{
 int t;
 
 cpureset();
 ulscanline=0;
 for (t=7; t>=0; t--)
 {
  if (slot[t].dll)
   slot[t].reset();
 }
 if (iie.iiemode)
 {
  iie.rdaux=iie.wraux=iie.lcaux=0;
  iie.cxoverride=0;
  iie.store80=0;
  iie.c3override=0x80;
 }
}

uint8_t MHz;

LARGE_INTEGER currenttime,throttlerate;
long long wantedtime;

/* Imprecise */
static void throttle (void)
{
 QueryPerformanceCounter(&currenttime);
 QueryPerformanceFrequency(&throttlerate);
 while (currenttime.QuadPart<wantedtime)
 {
  QueryPerformanceCounter(&currenttime);
  SwitchToThread();
 }
 wantedtime=currenttime.QuadPart+(throttlerate.QuadPart/(2*15720*MHz));
}

int8_t cpuloop(void)
{
 int x;
 
// printf("%04X- A=%02X X=%02X Y=%02X S=%02X P=%02X\n",cpugetpc(),cpugeta(),cpugetx(),cpugety(),cpugets(),cpugetp());
 
 ulscanline++;
 if (!fast6on) throttle();
 while (ulscanline>261)
 {
  ulscanline-=262;
 }
 
 x=ttyloop();
 if (x==-1) return 3;
 if (x==1) reset_all();
 if ((x>=10)&&(x<18))
 {
  if (slot[x-10].dll)
  {
   slot[x-10].config();
  }
 }
 return 0;
}

int slotassign (int slotno, char *module)
{
 int e;
 
 if ((slotno<0)||(slotno>7))
 {
  fprintf (stderr, ESTR_BADSLOT, slotno);
  return -1;
 }
 
 slot[slotno].dll=LoadLibrary(module);
 if (!(slot[slotno].dll))
 {
  fprintf(stderr, ESTR_OPEN, module);
  return -1;
 }

 if ((!(slot[slotno].mod_type=(void*)GetProcAddress(slot[slotno].dll,MOD_TYPE)))||
     (!(slot[slotno].mod_name=(void*)GetProcAddress(slot[slotno].dll,MOD_NAME)))||
     (!(slot[slotno].mod_pref=(void*)GetProcAddress(slot[slotno].dll,MOD_PREF)))||
     (!(slot[slotno].mod_vers=(void*)GetProcAddress(slot[slotno].dll,MOD_VERS)))||
     (!(slot[slotno].mod_copr=(void*)GetProcAddress(slot[slotno].dll,MOD_COPR)))) notamod(module);
 
 if ((*(slot[slotno].mod_type))!=0x544F4C53)
 {
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  fprintf (stderr, ESTR_ISNOT, module);
  fprintf (stderr, ESTR_NOTSLOT);
  return -1;
 }
 
 SFUNCTINO(slotno, preload_name, "preload_name");
 SFUNCTINO(slotno, preload_size, "preload_size");
 SFUNCTINO(slotno, preload_ptr,  "preload_ptr");

 SFUNCTINO(slotno, write16, "write16");
 SFUNCTINO(slotno, write256, "write256");
 SFUNCTINO(slotno, write2048, "write2048");

 SFUNCTINO(slotno, read16, "read16");
 SFUNCTINO(slotno, read256, "read256");
 SFUNCTINO(slotno, read2048, "read2048");

 SFUNCTINO(slotno, override_rd, "override_rd");
 SFUNCTINO(slotno, override_wr, "override_wr");
 
 SFUNCTINO(slotno, reset, "reset");
 SFUNCTINO(slotno, init, "init");
 SFUNCTINO(slotno, deinit, "deinit");

 SFUNCTINO(slotno, config, "config");
 
 if (*(slot[slotno].preload_ptr))
 {
  if (bload(*(slot[slotno].preload_name),*(slot[slotno].preload_size),*(slot[slotno].preload_ptr)))
  {
   FreeLibrary(slot[slotno].dll);
   slot[slotno].dll=0;
   return -1;
  }
 }
 
 e=slot[slotno].init();
 if (e==6502) /* plugin needs dma */
 {
  SFUNCTINO(slotno, dmainit, "dmainit");
  slot[slotno].dmainit(cpuwrite, cpuread);
 }
 else if (e)
 {
  fprintf (stderr, ESTR_INITMOD, module);
  FreeLibrary(slot[slotno].dll);
  slot[slotno].dll=0;
  return -1;
 }
 
 printf (ESTR_SLOT, slotno);
 printf (ESTR_IS,*(slot[slotno].mod_name),*(slot[slotno].mod_vers),*(slot[slotno].mod_copr));
 return 0;
}

int WINAPI WinMain(HINSTANCE ghinstance, HINSTANCE dummy, LPSTR cmdline, int nCmdShow)
{
 char inifile[1024];
 char modbuf[256], slotstr[6];
 char ttymod[256], cpumod[256];
 int s,l;
 int gotrom;
 
 hinstance=ghinstance;
 
 gotrom=0;
 
 getcwd(inifile,1010);
 strcat(inifile,"\\modapple.ini");
 
 for (l=0; l<8; l++) slot[l].dll=0;
 
 printf ("%s\n", verstr);
 
 l=*ttymod=*cpumod=0;
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CONSOLE, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(ttymod, modbuf);
 }
 
 l=GetPrivateProfileString(INI_SYSTEM, INI_CPU, 0, modbuf, 256, inifile);
 if (l)
 {
  strcpy(cpumod, modbuf);
 }
 
 for (s=0; s<8; s++)
 {
  sprintf (slotstr, "Slot%d", s);
  l=GetPrivateProfileString(INI_SLOTCARDS, slotstr, 0, modbuf, 256, inifile);
  if (l)
  {
   if (slotassign(s,modbuf))
    fprintf (stderr, ESTR_LINK, modbuf, s);
  }
 }
 
 if (!*ttymod)
 {
  fprintf (stderr, ESTR_NOTTY);
  return -1;
 }
 
 if (!*cpumod)
 {
  fprintf (stderr, ESTR_NOCPU);
  return -1;
 }
 
 mod_tty=0;
 mod_cpu=0;
 
 mod_tty=LoadLibrary(ttymod);
 if (!mod_tty)
 {
  fprintf(stderr, ESTR_OPEN, ttymod);
  return -1;
 }
 if ((!(tty_mod_type=(void*)GetProcAddress(mod_tty,MOD_TYPE)))||
     (!(tty_mod_name=(void*)GetProcAddress(mod_tty,MOD_NAME)))||
     (!(tty_mod_pref=(void*)GetProcAddress(mod_tty,MOD_PREF)))||
     (!(tty_mod_vers=(void*)GetProcAddress(mod_tty,MOD_VERS)))||
     (!(tty_mod_copr=(void*)GetProcAddress(mod_tty,MOD_COPR)))) notamod(ttymod);
 
 if ((*tty_mod_type)!=0x50595454)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, ttymod);
  fprintf (stderr, ESTR_NOTTTY);
  return -1;
 }
 
 TFUNCTINO(ttykey,"ttykey");
 TFUNCTINO(ttywrite,"ttywrite");
 TFUNCTINO(ttyread,"ttyread");
 TFUNCTINO(ttyout,"ttyout");
 TFUNCTINO(ttyin,"ttyin");
 TFUNCTINO(ttydeinit,"ttydeinit");
 TFUNCTINO(ttyinit,"ttyinit");
 TFUNCTINO(ttyloop,"ttyloop");
 
 mod_cpu=LoadLibrary(cpumod);
 if (!mod_cpu)
 {
  fprintf(stderr, ESTR_OPEN, cpumod);
  return -1;
 }
 if ((!(cpu_mod_type=(void*)GetProcAddress(mod_cpu,MOD_TYPE)))||
     (!(cpu_mod_name=(void*)GetProcAddress(mod_cpu,MOD_NAME)))||
     (!(cpu_mod_pref=(void*)GetProcAddress(mod_cpu,MOD_PREF)))||
     (!(cpu_mod_vers=(void*)GetProcAddress(mod_cpu,MOD_VERS)))||
     (!(cpu_mod_copr=(void*)GetProcAddress(mod_cpu,MOD_COPR)))) notamod(cpumod);
 
 if ((*cpu_mod_type)!=0x32303536)
 {
  canlibs();
  fprintf (stderr, ESTR_ISNOT, cpumod);
  fprintf (stderr, ESTR_NOTCPU);
  return -1;
 }
 
 CFUNCTINO(cpureset,"cpureset");
 CFUNCTINO(cpurun,"cpurun");
 CFUNCTINO(cpuinit,"cpuinit");
 CFUNCTINO(cpuseta,"cpuseta");
 CFUNCTINO(cpusetx,"cpusetx");
 CFUNCTINO(cpusety,"cpusety");
 CFUNCTINO(cpusets,"cpusets");
 CFUNCTINO(cpusetp,"cpusetp");
 CFUNCTINO(cpusetpc,"cpusetpc");
 CFUNCTINO(cpusetrate,"cpusetrate");
 CFUNCTINO(cpugeta,"cpugeta");
 CFUNCTINO(cpugetx,"cpugetx");
 CFUNCTINO(cpugety,"cpugety");
 CFUNCTINO(cpugets,"cpugets");
 CFUNCTINO(cpugetp,"cpugetp");
 CFUNCTINO(cpugetpc,"cpugetpc");
 CFUNCTINO(cpugetrate,"cpugetrate");

 l=GetPrivateProfileString(INI_SYSTEM, INI_BIOS, INI_DEF_ROM, modbuf, 256, inifile);

 iie.iiemode=0;
 
 if (!(bload(modbuf,12288,rom))) 
  gotrom=1;
 else
  if (!(bload(modbuf,16384,rom)))
  {
   iie.iiemode=1;
   gotrom=1;
  }
 if (!gotrom) fprintf (stderr, ESTR_NOBIOS);

 printf (ESTR_CPU);
 printf (ESTR_IS, *cpu_mod_name, *cpu_mod_vers, *cpu_mod_copr);
 printf (ESTR_TTY);
 printf (ESTR_IS, *tty_mod_name, *tty_mod_vers, *tty_mod_copr);

 holdc8=0;
 
 cpuinit(cpuwrite, cpuread, cpuloop);

 cpureset();
 cpusetrate(65);
// cpusetrate(1);
 
 MHz=1;
 wantedtime=0;
 fast6on=0;
 fast6=1;
 
 *ttykey=nCmdShow;
 ttyinit(&iie);
 reset_all();
 cpurun();
 ttydeinit();
 for (l=7; l>=0; l--) if (slot[l].dll) slot[l].deinit();
 
 canlibs();
 
 return 0;
}
